(			
	local	 _user32
	local	 IMaterialsLogStream = #()
	local	 IMaterialsSignatureTable = #()
	local	 ITextureSignatureTable = #()
	local	 IMaterialsIdenticalClusters = #()
	local	 ITexturesIdenticalClusters = #()
	local	 IMaterialsOriginalCount = 0
	local	 ITextureOriginalCount = 0	
	
	local instantiateIdenticalMaterials
	local instantiateMaterials
	local instantiateTextures
	local CreateUser32Assembly
	local createMaterialClusters
	local createTextureClusters
	local getMaterialSignature
	local getVRayMtlSignature
	local getTextureSignature
	local createIMaterialsSignatureTable
	local createITextureSignatureTable
	local sortBySignature
	local outputLog
	
	rollout rlt_log "Instantiate materials & textures" width:440 height:300
	(
		edittext edt_log "" pos:[10,10] width:340 height:280		
		button btn_start "Start" pos:[358,240] width:75 height:51
		checkbox chk_textures "Textures" pos:[363,23] width:66 height:17 checked:true		
		checkbox chk_materials "Materials" pos:[363,41] width:66 height:17 checked:true
		GroupBox grp1 "Instantiate" pos:[356,5] width:79 height:115
		label lbl_version "Version 1.0" pos:[365,65] width:59 height:16 enabled:true
		
		on btn_start pressed do
			instantiateIdenticalMaterials materials:chk_materials.state textures:chk_textures.state
		on btn_reportBug pressed  do
			reportBug()
	)
	
	-- Main
	fn instantiateIdenticalMaterials materials:true textures:true =
	(		
		CreateUser32Assembly()

		outputLog rlt_log "Starting..."
		outputLog rlt_log "\n"
		
		if textures do
		(
			createITextureSignatureTable rlt_log					

			createTextureClusters rlt_log
			
			outputLog rlt_log "Instantiating..."	
			outputLog rlt_log "\n"
			
			instantiateTextures rlt_log
			
			outputLog rlt_log ("Done ! Textures before/after Instantiation : " + (ITextureOriginalCount as string) + "/" + ((getclassinstances bitmaptexture).count as string))
			outputLog rlt_log "\n\n"
			outputLog rlt_log "\n"
		)

		if materials do
		(
			createIMaterialsSignatureTable rlt_log		

			createMaterialClusters rlt_log
			
			outputLog rlt_log "Instantiating...\n"
			outputLog rlt_log "\n"
			instantiateMaterials rlt_log

			outputLog rlt_log ("Done ! Materials before/after Instantiation : " + (IMaterialsOriginalCount as string) + "/" + ((getclassinstances VRayMtl).count as string) + "\n")
		)
	)
	
	/*  Instancie les matriaux. Rcupre d'abord tous les multimatriaux dans lesquels est rfrence le matriau  instancier,
	puis le remplace par le premier matriau du cluster, qui est le matriau de rfrence	*/	
	fn instantiateMaterials rlt =
	(		
		for i = 1 to IMaterialsIdenticalClusters.count do
		(
			cluster = IMaterialsIdenticalClusters[i]

			_user32.RedrawAllWindow (windows.getchildhwnd 0  rlt.title)[1]
			outputLog rlt ("Instantiating materials.............." +  (((i as float/IMaterialsIdenticalClusters.count)*100.) as integer) as string +"%\n" ) replaceText:true

			-- On zappe le premier car il reprsente l'instance originale
			for i = 2 to cluster.count do
			(
				inst = cluster[i][1]
				
				-- On rcupre tous les multimatriaux qui rfrencent le matriau  instancier
				dependents = refs.dependents inst

				for dep in dependents where (classof dep == multimaterial) do
					for j = 1 to dep.materialList.count where (isKindOf dep[j] VRayMtl) and (dep.materialList[j] == cluster[i][1]) do dep.materialList[j] = cluster[1][1]					

				for dep in dependents where (isValidNode dep) do
					if dep.mat == cluster[i][1] then dep.mat = cluster[1][1] 						
			)
		)
	)
	
	/*  Instancie les matriaux. Rcupre d'abord tous les multimatriaux dans lesquels est rfrence le matriau  instancier,
	puis le remplace par le premier matriau du cluster, qui est le matriau de rfrence
	*/	
	fn instantiateTextures rlt =
	(		
		for i = 1 to ITexturesIdenticalClusters.count do
		(
			cluster = ITexturesIdenticalClusters[i]

			_user32.RedrawAllWindow (windows.getchildhwnd 0  rlt.title)[1]
			outputLog rlt ("Instantiating textures.............." +  (((i as float/ITexturesIdenticalClusters.count)*100.) as integer) as string +"%\n" ) replaceText:true

			-- On zappe le premier car il reprsente l'instance originale
			for i = 2 to cluster.count do
			(
				inst = cluster[i][1]
				
				-- On rcupre tous les matriaux qui rfrencent la texture  instancier
				dependents = refs.dependents inst

				for j = 1 to dependents.count where isKindOf dep VRayMtl do
					for k = 1 to dep.maps.count where dep.maps[j] == cluster[i][1] do dependents[j].maps[k] = cluster[1][1]
			)
		)
	)
	
	-- From: Denis Trofimov
	fn CreateUser32Assembly =
	(
	source = "using System;\n"
	source += "using System.Runtime.InteropServices;\n"
	source += "class User32\n"
	source += "{\n"
	source += "	[DllImport(\"user32.dll\")]\n"
	source += "	static extern bool RedrawWindow(IntPtr hWnd, IntPtr lprcUpdate, IntPtr hrgnUpdate, uint flags);\n"
	source += "	static uint FULL_REDRAW = 0x0185;\n"
	source += "	public bool RedrawAllWindow(Int32 hWnd) { return RedrawWindow((IntPtr)hWnd, IntPtr.Zero, IntPtr.Zero, FULL_REDRAW); }\n"
	source += "}\n"

		csharpProvider = dotnetobject "Microsoft.CSharp.CSharpCodeProvider"
		compilerParams = dotnetobject "System.CodeDom.Compiler.CompilerParameters"

		compilerParams.ReferencedAssemblies.AddRange #("System.dll")

		compilerParams.GenerateInMemory = on
		compilerResults = csharpProvider.CompileAssemblyFromSource compilerParams #(source)

		User32Assembly = compilerResults.compiledAssembly
		_user32 = User32Assembly.CreateInstance "User32"
	)
	
	/* A partir de la table des signatures, cre des groupes de matriau partageant la mme signature */
	fn createMaterialClusters rlt =
	(
		scannedSignatures = #()
		c = IMaterialsSignatureTable.count
		identicalMaterials = 0.
		
		for i = 1 to IMaterialsSignatureTable.count do
		(
			_user32.RedrawAllWindow (windows.getchildhwnd 0  rlt.title)[1]
			outputLog rlt ("Creating material clusters.........." +  (((i as float/c as float)*100.) as integer) as string +"%\n" ) replaceText:true
			
			if (findItem scannedSignatures IMaterialsSignatureTable[i][2] == 0) do
			(			
				mat = IMaterialsSignatureTable[i]
				currentCluster = #()
				
				append currentCluster mat	
				
				for j = 1 to IMaterialsSignatureTable.count do
				(
					sMat = IMaterialsSignatureTable[j]
					if mat[2] == sMat[2] and mat[1] != sMat[1] do 
					(
						append currentCluster sMAT				
						identicalMaterials += 1.					
					)
				)
				
				if currentCluster.count > 1 do
					append IMaterialsIdenticalClusters currentCluster
				
				append scannedSignatures mat[2]
			)
		)

		_user32.RedrawAllWindow (windows.getchildhwnd 0  rlt.title)[1]
		outputLog rlt ("Identical Materials :.........." +  ((identicalMaterials/c)*100.) as string +"%\n" )
		print IMaterialsIdenticalClusters
	)
	
	/* A partir de la table des signatures, cre des groupes de matriau partageant la mme signature */
	fn createTextureClusters rlt =
	(
		scannedSignatures = #()
		c = ITextureSignatureTable.count
		identicalTextures = 0.
		
		for i = 1 to ITextureSignatureTable.count do
		(
			_user32.RedrawAllWindow (windows.getchildhwnd 0  rlt.title)[1]
			outputLog rlt ("Creating texture clusters.........." +  (((i as float/c as float)*100.) as integer) as string +"%\n" ) replaceText:true
			
			if (findItem scannedSignatures ITextureSignatureTable[i][2] == 0) do
			(			
				mat = ITextureSignatureTable[i]
				currentCluster = #()
				
				append currentCluster mat	
				
				for j = 1 to ITextureSignatureTable.count do
				(
					sMat = ITextureSignatureTable[j]
					if mat[2] == sMat[2] and mat[1] != sMat[1] do 
					(
						append currentCluster sMAT				
						identicalTextures += 1.					
					)
				)
				
				if currentCluster.count > 1 do
					append ITexturesIdenticalClusters currentCluster
				
				append scannedSignatures mat[2]
			)
		)

		_user32.RedrawAllWindow (windows.getchildhwnd 0  rlt.title)[1]
		outputLog rlt ("Identical Textures :.........." +  ((identicalTextures/c)*100.) as string +"%\n" )
		print IMaterialsIdenticalClusters
	)
	
	-- Calcule une signature unique du materiau en combinant maps et valeurs numriques
	fn getMaterialSignature mat =
	(
		signature = ""
		
		signature += mat.shaderType as string    --: integer
		signature += mat.wire as string    --: boolean
		signature += mat.twoSided as string    --: boolean
-- 		signature += mat.faceMap as string    --: boolean
-- 		signature += mat.faceted as string    --: boolean
-- 		signature += mat.shaderByName as string    --: string
-- 		signature += mat.opacityType as string    --: integer
		signature += mat.opacity as string    --: percent
-- 		signature += mat.filterColor as string    --: RGB color
-- 		signature += mat.filterMap as string    --: texturemap
-- 		signature += mat.opacityFallOffType as string    --: integer
-- 		signature += mat.opacityFallOff as string    --: percent
-- 		signature += mat.ior as string    --: float
-- 		signature += mat.wireSize as string    --: float
-- 		signature += mat.wireUnits as string    --: integer
-- 		signature += mat.applyReflectionDimming as string    --: boolean
-- 		signature += mat.dimLevel as string    --: float
-- 		signature += mat.reflectionLevel as string    --: float
-- 		signature += mat.sampler as string    --: integer
-- 		signature += mat.samplerQuality as string    --: float
-- 		signature += mat.samplerEnable as string    --: boolean
-- 		signature += mat.samplerAdaptThreshold as string    --: float
-- 		signature += mat.samplerAdaptOn as string    --: boolean
-- 		signature += mat.subSampleTextureOn as string    --: boolean
-- 		signature += mat.samplerAdvancedOptions as string    --: boolean
-- 		signature += mat.samplerByName as string    --: string
-- 		signature += mat.UserParam0 as string    --: float
-- 		signature += mat.UserParam1 as string    --: float
-- 		signature += mat.samplerUseGlobal as string    --: boolean

-- 		signature += mat.adTextureLock as string    --: boolean 
-- 		signature += mat.noExposureControl as string    --: boolean
-- 		signature += mat.exposureControlInvertSelfIllum as string    --: boolean
-- 		signature += mat.exposureControlInvertReflection as string    --: boolean
-- 		signature += mat.exposureControlInvertRefraction as string    --: boolean
		signature += mat.ambient as string    --: RGB color
		signature += mat.diffuse as string    --: RGB color
		signature += mat.specular as string    --: RGB color
-- 		signature += mat.adTextureLock as string    --: boolean
-- 		signature += mat.adLock as string    --: boolean
-- 		signature += mat.dsLock as string    --: boolean
		signature += mat.useSelfIllumColor as string    --: boolean
		signature += mat.selfIllumAmount as string    --: percent
		signature += mat.selfIllumColor as string    --: RGB color
		signature += mat.specularLevel as string    --: percent
		signature += mat.glossiness as string    --: percent
-- 		signature += mat.soften as string    --: float		

	 	signature += mat.mapEnables as string --: boolean array
	 	signature += mat.mapAmounts as string --: percent array		
		
		for map in mat.maps do
		(
			case classof map of
			(
				BitmapTexture:if map.filename != undefined then signature += (filenameFromPath map.filename) else signature += "none"
				default:signature += "none"				
			)
		)
		
		(gethashvalue signature 1)
	)
	
	fn getVRayMtlSignature mat =
	(
		signature = ""
	
		signature += mat.diffuse as string --: RGB color
		signature += mat.diffuse_roughness as string --: float
		signature += mat.reflection as string --: RGB color
		signature += mat.reflection_glossiness as string --: float
		--signature += mat.hilight_glossiness as string --: float
		signature += mat.reflection_subdivs as string --: integer
		signature += mat.reflection_fresnel as string --: boolean
		--signature += mat.reflection_maxDepth as string --: integer
		--signature += mat.reflection_exitColor (reflect_exitColor) as string --: RGB color
		--signature += mat.reflection_useInterpolation (useInterpolation) as string --: boolean
		--signature += mat.reflection_ior as string --: float
		--signature += mat.reflection_lockGlossiness as string --: boolean
		--signature += mat.reflection_lockIOR as string --: boolean
		--signature += mat.reflection_dimDistance as string --: worldUnits
		--signature += mat.reflection_dimDistance_on as string --: boolean
		--signature += mat.reflection_dimDistance_falloff as string --: float
		--signature += mat.reflection_affectAlpha as string --: integer
		--signature += mat.refraction as string --: RGB color
		--signature += mat.refraction_glossiness as string --: float
		--signature += mat.refraction_subdivs as string --: integer
		--signature += mat.refraction_ior as string --: float
		--signature += mat.refraction_fogColor as string --: RGB color
		--signature += mat.refraction_fogMult (refraction_fogMultiplier) as string --: float
		--signature += mat.refraction_fogBias as string --: float
		--signature += mat.refraction_affectShadows as string --: boolean
		--signature += mat.refraction_affectAlpha as string --: integer
		--signature += mat.refraction_maxDepth as string --: integer
		--signature += mat.refraction_exitColor (refract_exitColor) as string --: RGB color
		--signature += mat.refraction_useExitColor (refract_useExitColor) as string --: boolean
		--signature += mat.refraction_useInterpolation as string --: boolean
		--signature += mat.refraction_dispersion (refract_dispersion) as string --: float
		--signature += mat.refraction_dispersion_on (refract_dispersion_on) as string --: boolean
		--signature += mat.translucency_on (translucency_type) as string --: integer
		--signature += mat.translucency_thickness as string --: worldUnits
		--signature += mat.translucency_scatterCoeff as string --: float
		--signature += mat.translucency_fbCoeff as string --: float
		--signature += mat.translucency_multiplier as string --: float
		--signature += mat.translucency_color (translucent_color) as string --: RGB color
		--signature += mat.brdf_type as string --: integer
		--signature += mat.anisotropy (brdf_anisotropy) as string --: float
		--signature += mat.anisotropy_rotation (brdf_anisotropy_rotation) as string --: float
		--signature += mat.anisotropy_derivation (brdf_anisotropy_derivation) as string --: integer
		--signature += mat.anisotropy_axis (brdf_anisotropy_axis) as string --: integer
		--signature += mat.anisotropy_channel (brdf_anisotropy_channel) as string --: integer
		--signature += mat.soften (brdf_soften) as string --: float
		--signature += mat.brdf_fixDarkEdges as string --: boolean
		--signature += mat.option_traceDiffuse as string --: boolean
		--signature += mat.option_traceReflection as string --: boolean
		--signature += mat.option_traceRefraction as string --: boolean
		--signature += mat.option_doubleSided as string --: boolean
		--signature += mat.option_reflectOnBack as string --: boolean
		--signature += mat.option_useIrradMap as string --: boolean
		--signature += mat.refraction_fogUnitsScale_on (refraction_fogUnitScale_on) as string --: boolean
		--signature += mat.option_traceDiffuseAndGlossy as string --: integer
		--signature += mat.option_cutOff as string --: float
		--signature += mat.preservationMode as string --: integer
		--signature += mat.option_environment_priority as string --: integer
		--signature += mat.effect_id as string --: integer
		--signature += mat.override_effect_id as string --: boolean
		--signature += mat.option_clampTextures as string --: boolean
		--signature += mat.texmap_diffuse as string --: texturemap
		--signature += mat.texmap_diffuse_on as string --: boolean
		--signature += mat.texmap_diffuse_multiplier as string --: float
		--signature += mat.texmap_reflection as string --: texturemap
		--signature += mat.texmap_reflection_on as string --: boolean
		--signature += mat.texmap_reflection_multiplier as string --: float
		--signature += mat.texmap_refraction as string --: texturemap
		--signature += mat.texmap_refraction_on as string --: boolean
		--signature += mat.texmap_refraction_multiplier as string --: float
		--signature += mat.texmap_bump as string --: texturemap
		--signature += mat.texmap_bump_on as string --: boolean
		--signature += mat.texmap_bump_multiplier as string --: float
		--signature += mat.texmap_reflectionGlossiness as string --: texturemap
		--signature += mat.texmap_reflectionGlossiness_on as string --: boolean
		--signature += mat.texmap_reflectionGlossiness_multiplier as string --: float
		--signature += mat.texmap_refractionGlossiness as string --: texturemap
		--signature += mat.texmap_refractionGlossiness_on as string --: boolean
		--signature += mat.texmap_refractionGlossiness_multiplier as string --: float
		--signature += mat.texmap_refractionIOR (texmap_ior) as string --: texturemap
		--signature += mat.texmap_refractionIOR_on (texmap_ior_on) as string --: boolean
		--signature += mat.texmap_refractionIOR_multiplier (texmap_ior_multiplier) as string --: float
		--signature += mat.texmap_displacement as string --: texturemap
		--signature += mat.texmap_displacement_on as string --: boolean
		--signature += mat.texmap_displacement_multiplier as string --: float
		--signature += mat.texmap_translucent as string --: texturemap
		--signature += mat.texmap_translucent_on as string --: boolean
		--signature += mat.texmap_translucent_multiplier as string --: float
		--signature += mat.texmap_environment as string --: texturemap
		--signature += mat.texmap_environment_on as string --: boolean
		--signature += mat.texmap_hilightGlossiness as string --: texturemap
		--signature += mat.texmap_hilightGlossiness_on as string --: boolean
		--signature += mat.texmap_hilightGlossiness_multiplier as string --: float
		--signature += mat.texmap_reflectionIOR as string --: texturemap
		--signature += mat.texmap_reflectionIOR_on as string --: boolean
		--signature += mat.texmap_reflectionIOR_multiplier as string --: float
		--signature += mat.texmap_opacity as string --: texturemap
		--signature += mat.texmap_opacity_on as string --: boolean
		--signature += mat.texmap_opacity_multiplier as string --: float
		--signature += mat.texmap_roughness as string --: texturemap
		--signature += mat.texmap_roughness_on as string --: boolean
		--signature += mat.texmap_roughness_multiplier as string --: float
		--signature += mat.texmap_anisotropy as string --: texturemap
		--signature += mat.texmap_anisotropy_on as string --: boolean
		--signature += mat.texmap_anisotropy_multiplier as string --: float
		--signature += mat.texmap_anisotropy_rotation as string --: texturemap
		--signature += mat.texmap_anisotropy_rotation_on as string --: boolean
		--signature += mat.texmap_anisotropy_rotation_multiplier as string --: float
		--signature += mat.reflect_minRate as string --: integer
		--signature += mat.reflect_maxRate as string --: integer
		--signature += mat.reflect_interpSamples as string --: integer
		--signature += mat.reflect_colorThreshold (reflect_clrThresh) as string --: float
		--signature += mat.reflect_normalThreshold (reflect_nrmThresh) as string --: float
		--signature += mat.refract_minRate as string --: integer
		--signature += mat.refract_maxRate as string --: integer
		--signature += mat.refract_interpSamples as string --: integer
		--signature += mat.refract_colorThreshold (refract_clrThresh) as string --: float
		--signature += mat.refract_normalThreshold (refract_nrmThresh) as string --: float
		
		case classOf mat of
		(
			standard:
			(
				for map in mat.maps do
				(
					case classof map of
					(
						BitmapTexture:if map.filename != undefined then signature += (filenameFromPath map.filename) else signature += "none"
						default:signature += "none"				
					)
				)
			)
			VRayMtl:
			(
				mapNames = for pn in (getPropNames mat) where (matchPattern pn pattern:"*texmap_*" and not matchPattern pn pattern:"*_on" and not matchPattern pn pattern:"*_multiplier") collect pn
				maps = for nm in mapNames collect getProperty mat nm
				for map in maps do
				(
					case classof map of
					(
						BitmapTexture:if map.filename != undefined then signature += (filenameFromPath map.filename) else signature += "none"
						default:signature += "none"				
					)
				)
			)
		)
		
		(gethashvalue signature 1)
	)
	
	-- Calcule une signature unique d'une texture en combinant maps et valeurs numriques
	fn getTextureSignature texture =
	(
		signature = ""
		
		signature += texture.clipu as string
		signature += texture.clipv as string
		signature += texture.clipw as string
		signature += texture.cliph as string
		signature += texture.monoOutput as string
		signature += texture.rgbOutput as string
		signature += texture.alphaSource as string
		signature += texture.preMultAlpha as string
-- 		signature += texture.bitmap as string
		signature += texture.coords as string

		texture.fileName
		
-- 		Coordonnes de texture
		signature += texture.coords.blur as string 
		signature += texture.coords.mapping as string
		signature += texture.coords.mapChannel as string 
		signature += texture.coords.mappingType as string 
		signature += texture.coords.UVW_Type as string 
		signature += texture.coords.U_Mirror as string 
		signature += texture.coords.V_Mirror as string 
		signature += texture.coords.U_Tile as string 
		signature += texture.coords.V_Tile as string 
		signature += texture.coords.showMapOnBack as string 
		signature += texture.coords.Noise_On as string 
		signature += texture.coords.Noise_Animate as string 
		signature += texture.coords.UVTransform as string
		signature += texture.coords.realWorldScale as string 
		signature += texture.coords.realWorldHeight as string 
		signature += texture.coords.realWorldWidth as string 
		signature += texture.coords.phase as string 
		signature += texture.coords.Blur_Offset as string 
-- 		signature += texture.coords.U_Offset as string 
-- 		signature += texture.coords.V_Offset as string 
-- 		signature += texture.coords.U_Tiling as string 
-- 		signature += texture.coords.V_Tiling as string 
-- 		signature += texture.coords.U_ as string 
-- 		signature += texture.coords.V_ as string 
-- 		signature += texture.coords.W_ as string 
		signature += texture.coords.Noise_Amount as string 
		signature += texture.coords.Noise_Size as string 
		signature += texture.coords.Noise_Levels as string 
		
		(gethashvalue signature 1)
	)
	
	-- Renvoie un tableau de tableaux de type #(materiau, signature) composes de rfrences  des materiaux et de leur signature
	fn createIMaterialsSignatureTable rlt =
	(
		"createIMaterialsSignatureTable"
		IMaterialsSignatureTable = #()
		
		mats = getclassinstances VRayMtl	
		IMaterialsOriginalCount = mats.count
		c = mats.count as float
		for i = 1 to  mats.count do 
		(
			append IMaterialsSignatureTable #(mats[i],(getVRayMtlSignature mats[i]))
			_user32.RedrawAllWindow (windows.getchildhwnd 0  rlt.title)[1]
			outputLog rlt ("Creating material signatures table.........." +  (((i/c)*100.) as integer) as string +"%\n" ) replaceText:true
		)
	)
	
	-- Renvoie un tableau de tableaux de type #(bitmaptexture, signature) composes de rfrences  des materiaux et de leur signature
	fn createITextureSignatureTable rlt =
	(
		ITextureSignatureTable = #()		
		textures = getclassinstances bitmaptexture	
		ITextureOriginalCount = textures.count
		c = textures.count as float
		for i = 1 to textures.count do
		(
			if not textures[i].coords.isanimated do
				append ITextureSignatureTable #(textures[i],(getTextureSignature textures[i]))
					
			_user32.RedrawAllWindow (windows.getchildhwnd 0  rlt.title)[1]
			outputLog rlt ("Creating texture signatures table.........." +  (((i/c)*100.) as integer) as string +"%\n" ) replaceText:true
		)
	)
	
	/* Log manager
	rlt est le rollout sur lequel crire le log
	s est la string  ajouter au log
	replaceText permet de remplacer le texte de la ligne.
	afin dviter une accumulation de textes identiques  (comme pour l'affichage d'une progression en pourcentage)
	*/
	fn outputLog rlt s replaceText:false =
	(
		if replaceText do deleteitem IMaterialsLogStream IMaterialsLogStream.count
		
		append IMaterialsLogStream s
		
		rlt.edt_log.text = ""
		for str in IMaterialsLogStream do		
			rlt.edt_log.text += str		
	)

	try (destroyDialog rlt_log)catch()
	CreateDialog rlt_log
)
